SystemOrganization addCategory: #Scheduler!

!Time class methodsFor: '*Scheduler-override' stamp: 'jrp 9/21/2004 20:11'!
now
	"Answer a Time representing the time right now - this is a 24 hour clock."

	^ self seconds: self totalSeconds \\ 86400 nanoSeconds: MillisecondClockOffset milliSeconds * NanosInMillisecond
! !

TestCase subclass: #SchedulerTest
	instanceVariableNames: 'scheduler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scheduler'!

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 7/14/2004 22:16'!
setUp

	scheduler _ TaskScheduler new.
	scheduler start.! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 7/14/2004 22:43'!
tearDown

	scheduler stop.! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:40'!
testDailyTaskDoesntGetsAheadIfRunExplicitly

	| task |

	task _ scheduler do: ['task ran'] at: DateAndTime now + 10 seconds.
	
	self should: [task nextRunTime asDate = Date today].
	
	task run. "run explicitly before scheduled time"
	
	self should: [task nextRunTime asDate = Date today].! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:33'!
testNewTaskScheduledToRunEarlierInDayGetsRescheduledLaterInDay

	| message when task |

	when _ (DateAndTime now - 1 minute) asTime.
	
	task _ scheduler do: [message _ 'task ran'] at: when.
	
	1.1 seconds asDelay wait.

	self should: [message isNil].
	self should: [scheduler taskCount = 1].
	self should: [task nextRunTime = (DateAndTime date: (Date today addDays: 1) time: when)]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 22:11'!
testRunAtSomeTimeInTheFuture

	| task timeToRun |
	timeToRun _ DateAndTime now + 2 second.
	task _ scheduler do: [] at: timeToRun every: 1 hour.
	
	2.1 seconds asDelay wait.
	
	self should: [task nextRunTime = (timeToRun + 1 hour)].
	self should: [task schedule schedule first = 1 hour].
	self should: [scheduler taskCount = 1]
	
	
! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:33'!
testRunDailyTask

	| message when task |

	when _ (DateAndTime now + 1 second) asTime.
	
	task _ scheduler do: [message _ 'task ran'] at: when.
	
	1.1 seconds asDelay wait.

	self should: [message = 'task ran'].
	self should: [scheduler taskCount = 1].
	self should: [task nextRunTime = (DateAndTime date: (Date today addDays: 1) time: when)]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 22:33'!
testRunDailyTaskNeverGetsBehind

	| task |

	scheduler stop.
	task _ scheduler do: ['task ran'] at: DateAndTime now.
	
	task startDateAndTime: (task nextRunTime - 2 days) duration: 1 day.
	scheduler start.
	1 second asDelay wait.
	
	self should: [task nextRunTime asDate = Date tomorrow].
	
	task startDateAndTime: ((DateAndTime now + 180 seconds) - 1 day) duration: 1 day.
	1 second asDelay wait.
	
	self should: [task nextRunTime asDate = Date today]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/4/2004 23:03'!
testRunEveryOtherHourOnTheHour

	| currentTime task |
	scheduler stop.
	currentTime _ DateAndTime now asTime.
	
	task _ scheduler do: [] at: currentTime hour asString every: 2 hours.
	task run.
		
	self should: [task nextRunTime asTime hour = ((currentTime addSeconds: 2 * 60 * 60) hour)].

	self should: [task nextRunTime > DateAndTime now]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/19/2004 19:46'!
testRunEveryQuarterSecond

	| runCount |
	runCount _ 0.

	scheduler do: [runCount _ runCount + 1] every: 250 milliSeconds.
	
	900 milliSeconds asDelay wait.
	
	self should: [runCount = 3].
	self should: [scheduler taskCount = 1]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:33'!
testRunEverySecond

	| runCount |
	runCount _ 0.

	scheduler do: [runCount _ runCount + 1] every: 1 second.
	
	3.1 seconds asDelay wait.
	
	self should: [runCount = 3].
	self should: [scheduler taskCount = 1]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:34'!
testRunInASecond

	| ranWhen now |
	now _ DateAndTime now.
	
	scheduler do: [ranWhen _ DateAndTime now] at: (now + 1 seconds).
	
	1.1 seconds asDelay wait.

	self should: [ranWhen = (now + 1 seconds)]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 11:25'!
testRunOnceInPastRunsOnce

	| message |
	scheduler doOnce: [message _ 'task ran'] at: (DateAndTime now - 1 second).
	
	1 second asDelay wait.
	
	self should: [message = 'task ran'].
	self should: [scheduler taskCount = 0]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:32'!
testRunOnlyOnceWithFullyQualifiedDateAndTime

	| runCount |
	runCount _ 0.
	scheduler doOnce: [runCount _ runCount + 1] at: (DateAndTime now + 1 seconds).
	
	1.1 seconds asDelay wait.
	
	self should: [runCount = 1].
	self should: [scheduler taskCount = 0]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:44'!
testRunOnlyOnceWithOnlyTimeSpecified

	| runCount |
	runCount _ 0.
	scheduler doOnce: [runCount _ runCount + 1] at: ((DateAndTime now + 1 seconds) asTime).
	
	1.1 seconds asDelay wait.
	
	self should: [runCount = 1].
	self should: [scheduler taskCount = 0]! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/5/2004 10:34'!
testRunSeveralTasks

	| task1 task2 now |
	now _ DateAndTime now.
	
	scheduler do: [task1 _ 'task 1'] at: (now + 1 second).
	scheduler do: [task2 _ 'task 2'] at: (now + 1 second).
	
	1.1 seconds asDelay wait.

	self should: [task1 = 'task 1'].
	self should: [task2 = 'task 2'].! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/6/2004 16:06'!
testRunUsingCustomSchedule

	| runCount schedule |
	runCount _ 0.
	
	schedule _ Schedule starting: DateAndTime now duration: 1 minute.
	schedule schedule: {1 second. 2 seconds. 20 seconds}.
	
	scheduler do: [runCount _ runCount + 1] following: schedule.
	3.1 seconds asDelay wait.
	
	self should: [runCount = 2].
	self should: [scheduler taskCount = 1]
! !

!SchedulerTest methodsFor: 'as yet unclassified' stamp: 'jrp 9/6/2004 16:37'!
testRunUsingCustomScheduleThatEnds

	| runCount schedule |
	runCount _ 0.
	
	schedule _ Schedule starting: (DateAndTime now + 1 second) duration: 4 seconds.
	schedule schedule: {1 second. 2 seconds}.
	
	scheduler do: [runCount _ runCount + 1] following: schedule.
	5.1 seconds asDelay wait.
	
	self should: [runCount = 3].
	self should: [scheduler taskCount = 0]
! !

Object subclass: #MillisecondClockOffset
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scheduler'!

MillisecondClockOffset class
	instanceVariableNames: 'milliSecondOffset'!

MillisecondClockOffset class
	instanceVariableNames: 'milliSecondOffset'!

!MillisecondClockOffset class methodsFor: 'as yet unclassified' stamp: 'jrp 9/19/2004 19:11'!
initialize

	super initialize.

	Smalltalk addToStartUpList: self.! !

!MillisecondClockOffset class methodsFor: 'as yet unclassified' stamp: 'jrp 9/19/2004 19:40'!
getMilliSecondOffset

   | lastSecond |
   lastSecond _ Time totalSeconds.
   [lastSecond = Time totalSeconds]
       whileTrue: [(Delay forMilliseconds: 1) wait].

   ^ Time millisecondClockValue \\ 1000.! !

!MillisecondClockOffset class methodsFor: 'as yet unclassified' stamp: 'jrp 9/21/2004 20:08'!
milliSeconds

	^ milliSecondOffset
		ifNil: [0]
		ifNotNil: [(Time millisecondClockValue - milliSecondOffset) \\ 1000]! !

!MillisecondClockOffset class methodsFor: 'as yet unclassified' stamp: 'jrp 9/21/2004 20:09'!
startUp: resuming

	resuming ifFalse: [^ self].
	
	milliSecondOffset _ nil.
	
	[milliSecondOffset _ self getMilliSecondOffset] fork! !

Object subclass: #ScheduledTask
	instanceVariableNames: 'description scheduler task runHistory workerProcess schedule nextRunTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scheduler'!

!ScheduledTask class methodsFor: 'as yet unclassified' stamp: 'jrp 7/15/2004 00:01'!
new

	^ super new initialize! !

!ScheduledTask methodsFor: 'private' stamp: 'jrp 9/5/2004 10:46'!
convertToTimeParameter: parameter

	^ [parameter asDateAndTime] on: Error do: [Time readFrom: parameter readStream]! !

!ScheduledTask methodsFor: 'executing' stamp: 'jrp 7/15/2004 10:20'!
delete

	scheduler removeTask: self! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/15/2004 13:04'!
description

	^ description! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/15/2004 13:04'!
description: aString

	description _ aString! !

!ScheduledTask methodsFor: 'initialization' stamp: 'jrp 9/5/2004 10:38'!
initialize

	runHistory _ OrderedCollection new! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/30/2004 08:06'!
isRunning

	workerProcess ifNil: [^ false].
	
	^ workerProcess isTerminated not! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/30/2004 08:25'!
isSuspended

	workerProcess ifNil: [^ false].
	
	^ workerProcess isTerminated not and: [workerProcess isSuspended]! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 9/5/2004 22:42'!
nextRunTime

	^ nextRunTime! !

!ScheduledTask methodsFor: 'printing' stamp: 'jrp 9/4/2004 23:21'!
printOn: aStream

	aStream nextPutAll: 'Scheduled Task'; cr.
	self description ifNotNil: [aStream nextPutAll: self description; cr].
	
	self isRunning ifTrue: [aStream nextPutAll: 'CURRENTLY RUNNING !!!!'].
	self isSuspended ifTrue: [aStream nextPutAll: ' (suspended)'].
	
	aStream
		cr;
		nextPutAll: 'Next Run Time : ', self nextRunTime asString;
		cr;
		nextPutAll: 'Last Run Time : ', (runHistory isEmpty 
			ifTrue: ['<never run>']
			ifFalse: [runHistory last asString]).! !

!ScheduledTask methodsFor: 'executing' stamp: 'jrp 7/30/2004 08:22'!
resume

	self isSuspended ifTrue: [workerProcess resume]! !

!ScheduledTask methodsFor: 'executing' stamp: 'cds 12/18/2007 09:16'!
run

	workerProcess _ [[[task value] ifError: []] ensure: [runHistory add: DateAndTime now]] fork.

	self wasRun! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/15/2004 13:04'!
runHistory

	^ runHistory! !

!ScheduledTask methodsFor: 'executing' stamp: 'jrp 9/4/2004 22:32'!
runIfNecessary

	((DateAndTime now >= self nextRunTime) and: [self isRunning not]) ifTrue: [self run]! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 9/5/2004 22:10'!
schedule

	^ schedule! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 9/6/2004 15:41'!
schedule: aSchedule

	schedule _ aSchedule.
	nextRunTime _ schedule start.
	
	self setNextRunTime
	! !

!ScheduledTask methodsFor: 'private' stamp: 'jrp 9/5/2004 22:38'!
scheduleHasOnlyZeroDuration

	^ schedule schedule size = 1 and: [schedule schedule first = Duration zero]! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/15/2004 13:04'!
scheduler

	^ scheduler! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/14/2004 23:34'!
scheduler: aTaskScheduler

	scheduler _ aTaskScheduler! !

!ScheduledTask methodsFor: 'private' stamp: 'jrp 9/6/2004 16:34'!
setNextRunTime

	self scheduleHasOnlyZeroDuration ifTrue: [^ self].
	
	schedule scheduleDo:
		[:each | (each > DateAndTime now) ifTrue: [nextRunTime _ each. ^ self]].
	
	self delete "delete if no more scheduled tasks"! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 9/5/2004 22:43'!
startDateAndTime: aDateAndTime duration: aDuration

	| timeParameter |
	timeParameter _ self convertToTimeParameter: aDateAndTime.
	 
	nextRunTime _ ((timeParameter isKindOf: Time)
		ifTrue:
			[DateAndTime date: Date today time: timeParameter]
		ifFalse:
			[timeParameter]).
			
	schedule _ Schedule starting: self nextRunTime duration: 1000 weeks.
	schedule schedule: {aDuration}.	
		
	self setNextRunTime

! !

!ScheduledTask methodsFor: 'executing' stamp: 'jrp 7/30/2004 08:19'!
stop

	self isRunning ifTrue: [workerProcess terminate]

	! !

!ScheduledTask methodsFor: 'executing' stamp: 'jrp 7/30/2004 08:20'!
suspend

	self isRunning ifTrue: [workerProcess suspend]

	! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/15/2004 13:04'!
task

	^ task! !

!ScheduledTask methodsFor: 'accessing' stamp: 'jrp 7/14/2004 23:32'!
task: aBlock

	task _ aBlock! !

!ScheduledTask methodsFor: 'executing' stamp: 'jrp 9/5/2004 22:39'!
wasRun

	self scheduleHasOnlyZeroDuration ifTrue: [self delete].
	
	self setNextRunTime! !

Object subclass: #TaskScheduler
	instanceVariableNames: 'running tasks accessProtect'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scheduler'!

!TaskScheduler class methodsFor: 'as yet unclassified' stamp: 'jrp 7/14/2004 22:43'!
new

	^ super new initialize! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 9/5/2004 10:48'!
addTask: aScheduledTask

	^ accessProtect critical: [tasks add: aScheduledTask]! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 9/4/2004 23:12'!
do: aBlock at: when

	^ self do: aBlock at: when every: 1 day! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 9/5/2004 22:36'!
do: aBlock at: when every: aDuration

	^ self addTask: (ScheduledTask new
		scheduler: self;
		task: aBlock;
		startDateAndTime: when
		duration: aDuration)! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 9/4/2004 23:24'!
do: aBlock every: aDuration

	^ self do: aBlock at: Time now every: aDuration! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 9/5/2004 22:52'!
do: aBlock following: aSchedule

	^ self addTask: (ScheduledTask new
		scheduler: self;
		task: aBlock;
		schedule: aSchedule)! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 9/5/2004 10:03'!
doOnce: aBlock at: when

	^ self do: aBlock at: when every: Duration zero! !

!TaskScheduler methodsFor: 'initialization' stamp: 'jrp 7/14/2004 23:06'!
initialize

	running _ false.
	tasks _ OrderedCollection new.
	accessProtect _ Semaphore forMutualExclusion! !

!TaskScheduler methodsFor: 'printing' stamp: 'jrp 5/14/2005 22:37'!
printOn: aStream

	aStream
		nextPutAll: 'Task scheduler is ', (running ifTrue: ['running'] ifFalse: ['stopped']);
		nextPutAll: ' with ';
		nextPutAll: self taskCount asWords;
		nextPutAll:' task'.
		
	self taskCount = 1 ifFalse: [aStream nextPut: $s]! !

!TaskScheduler methodsFor: 'scheduling' stamp: 'jrp 7/15/2004 08:36'!
removeTask: aScheduledTask

	accessProtect critical: [tasks remove: aScheduledTask]! !

!TaskScheduler methodsFor: 'executing' stamp: 'jrp 9/19/2004 19:51'!
start

	| process |
	running ifTrue: [^ self].
	
	running _ true.
	
	process _ [
		[running] whileTrue: 
			[0.25 seconds asDelay wait.
			self tasksDo: [:each | each runIfNecessary]]] newProcess.
	
	(process respondsTo: #name:) ifTrue: [process name: 'taskScheduler'].
	
	process priority: Processor systemBackgroundPriority.
	
	process resume.
! !

!TaskScheduler methodsFor: 'executing' stamp: 'jrp 7/14/2004 22:43'!
stop

	running _ false! !

!TaskScheduler methodsFor: 'accessing' stamp: 'jrp 7/15/2004 10:19'!
taskAt: aNumber

	^ accessProtect critical: [tasks at: aNumber]! !

!TaskScheduler methodsFor: 'accessing' stamp: 'jrp 7/15/2004 09:30'!
taskCount

	^ accessProtect critical: [tasks size]! !

!TaskScheduler methodsFor: 'accessing' stamp: 'jrp 7/15/2004 09:25'!
tasksDo: aOneArgBlock
	
	| copiedTasks |
	accessProtect critical: [copiedTasks _ tasks copy].
	
	copiedTasks do: aOneArgBlock! !

!DateAndTime class methodsFor: '*Scheduler-override' stamp: 'jrp 9/19/2004 19:26'!
now
	^ self basicNew 
		ticks: (Duration 
				days: SqueakEpoch 
				hours: 0 
				minutes: 0 
				seconds: self totalSeconds 
				nanoSeconds: MillisecondClockOffset milliSeconds * NanosInMillisecond) ticks
		offset: self localOffset;
		yourself
! !
MillisecondClockOffset initialize!
